/***************************************************************************\
  "lexer.h" - Lexical Analyzer (Phase B)
  Revision 5; 10 Nov 2002 - 24 Nov 2002
  Copyright (C) 2002  Kevin P. Albrecht

  References:
    MCD 2.1.5; 2.1.5.1
\***************************************************************************/

#ifndef LEXER_H
#define LEXER_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define LEXER_DIGITS "0123456789"
#define LEXER_NUMBER_ELEMENTS "0123456789."
#define LEXER_LETTERS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define LEXER_SPECIALS "+-*/\\=(),<>$.[]\":;{}&#"
#define LEXER_MIN_WORD_LEN 2
#define LEXER_MAX_WORD_LEN 6
#define LEXER_TAB_WIDTH 5

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdlib.h>
#include <math.h>

// Local //
#include "inputbuffer.h"
#include "global.h"
#include "token.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //
void lex_init ();
void lex_deinit ();
Token *lex_get (Token *oldToken);

// Internal //
bool mCheckForInclude ();
void mSkipBlankSpaces ();
void mCheckIndent ();
Token *mGetWord ();
Token *mGetNumber ();
Token *mGetString ();
Token *mGetEOL ();
Token *mGetSpecial ();
Token *mGetEOF ();
Token *mGetError ();

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// lex_init
//// Initialize Lexer Module
void lex_init ()
{
  // Blank right now //
}

//// lex_deinit
//// Deinitialize Lexer Module
void lex_deinit ()
{
  // Blank right now //
}

//// lex_get
//// Get next token from buffer, delete the oldToken
Token *lex_get (Token *oldToken)
{
  char c;

  // This allows the current line to be incremented in certain conditions //
  if (oldToken) if (oldToken->code == codeEOL) gnCurrentLine++;

  // Find indentation //
  if (oldToken) if ((oldToken->code == codeEOL) || (oldToken->code == codeStart))
  {
    mCheckIndent ();
  }

  // Delete the old token //
  if (oldToken) tok_del (oldToken);

  mSkipBlankSpaces ();
  if (mCheckForInclude() == true)
  {
    mCheckIndent ();
  }
  mSkipBlankSpaces ();
  c = ib_get ();

  if (k_inSet (c, LEXER_LETTERS))       return mGetWord ();
  else if (k_inSet (c, LEXER_DIGITS))   return mGetNumber ();
  else if (c == '\"')                   return mGetString ();
  else if (c == 0)                      return mGetEOL ();
  else if (k_inSet (c, LEXER_SPECIALS)) return mGetSpecial ();
  else if (c == INPUTBUFFER_EOF)        return mGetEOF ();
  else                                  return mGetError ();

  // This would allow ';' for EOLs:
  // else if (c == 0) || (c == ';'))    return mGetEOL ();
}

//// mCheckForInclude
//// Check for inlusion symbol '$'; return true if found
bool mCheckForInclude ()
{
  char *filename;
  int pos;

  if (ib_get() == '$')
  {
    filename = k_new (256);
    k_copy (filename, "");
    pos = 0;

    ib_seek ();
    while (ib_get() != 0)
    {
      filename[pos++] = ib_get();
      ib_seek ();
    }

    filename[pos - 1] = '\0';

    ib_include (filename);

    k_del (filename);

    return true;
  }

  return false;
}

//// mSkipBlankSpaces
//// Remove blank spaces from buffer; includes comments, and all whitespace
void mSkipBlankSpaces ()
{
  int nLine;

  while (((ib_get() <= ' ') && (ib_get() > 0)) || (ib_get() == '{'))
  {
    if (ib_get() == '{')
    {
      ib_seek ();
      if (ib_get() == '{') // single line comment
      {
        while (ib_get() != 0)
        {
          ib_seek ();
        }
        if (ib_get() == 0)
        {
          // DEBUGGED 10/29/02 //
          // By returning now, lex_get() can return an EOL token //
          return;
        }
      }
      else // multiple line comment
      {
        nLine = gnCurrentLine;
        while (ib_get() != '}')
        {
          if (ib_get() == INPUTBUFFER_EOF)
          {
            gnCurrentLine = nLine;
            err (true, EF_INCOMPLETE_COMMENT, NULL);
            erre (EFF_TERMINATE);
          }
          ib_seek ();
        }
      }
    }
    ib_seek ();
  }
}

//// mCheckIndent
//// Check for indentation
void mCheckIndent ()
{
  int nSpaces = 0, nLine = 0;

  while (((ib_get() <= ' ') && (ib_get() > 0)) || (ib_get() == '{'))
  {
    if (ib_get() == '{') // comment
    {
      nSpaces++;
      ib_seek ();

      if (ib_get() == '{') // single line comment
      {
        while (ib_get() != 0)
        {
          nSpaces++;
          ib_seek ();
        }
        if (ib_get() == 0)
        {
          // break here since the end of line is reached //
          break;
        }
      }
      else // multiple line comment
      {
        nLine = gnCurrentLine;
        while (ib_get() != '}')
        {
          if (ib_get() == INPUTBUFFER_EOF)
          {
            gnCurrentLine = nLine;
            err (true, EF_INCOMPLETE_COMMENT, NULL);
            erre (EFF_TERMINATE);
          }
          nSpaces++;
          ib_seek ();
        }
        nSpaces++;
        ib_seek();
      }
    }
    else if (ib_get() == '\t') // tab
    {
      nSpaces += LEXER_TAB_WIDTH;
      ib_seek ();
    }
    else // other whitespace
    {
      nSpaces++;
      ib_seek ();
    }
  }

  // ensure the line is not empty //
  if (ib_get() != 0)
  {
    if (nSpaces < gnIndent) scope_end ();
    else if (nSpaces > gnIndent) scope_begin ();
    gnIndent = nSpaces;
  }
}

//// mGetWord
//// Return a word from the buffer
Token *mGetWord ()
{
  // Vars //
  Token *token;
  char c;
  int nLexemePos;
  int nLen;

  token = tok_new();
  c = ib_get ();
  nLexemePos = 0;

  // Put the characters of the token into 'token->lexeme' //
  do {
    token->lexeme[nLexemePos++] = c;
    ib_seek();
    c = ib_get ();
  } while (k_inSet (c, LEXER_LETTERS) || k_inSet (c, LEXER_DIGITS));
  token->lexeme[nLexemePos] = '\0';

  // Length of the token //
  nLen = strlen (token->lexeme);

  // Assume it's an identifier //
  token->code = codeIdentifier;

  // Verify length of lexeme //
  if ((nLen >= LEXER_MIN_WORD_LEN) && (nLen <= LEXER_MAX_WORD_LEN))
  {
    switch (nLen)
    {
      case 2:
        if (strcmp (token->lexeme, "or") == 0) token->code = codeOR;
        break;
      case 3:
        if (strcmp (token->lexeme, "and") == 0) token->code = codeAND;
        else if (strcmp (token->lexeme, "def") == 0) token->code = codeDEF;
        else if (strcmp (token->lexeme, "int") == 0) token->code = codeINT;
        else if (strcmp (token->lexeme, "rat") == 0) token->code = codeRAT;
        else if (strcmp (token->lexeme, "str") == 0) token->code = codeSTR;
        break;
      case 4:
        if (strcmp (token->lexeme, "file") == 0) token->code = codeFILE;
        else if (strcmp (token->lexeme, "goto") == 0) token->code = codeGOTO;
        else if (strcmp (token->lexeme, "cond") == 0) token->code = codeCOND;
        else if (strcmp (token->lexeme, "loop") == 0) token->code = codeLOOP;
        break;
      case 5:
        if (strcmp (token->lexeme, "print") == 0) token->code = codePRINT;
        break;
      case 6:
        if (strcmp (token->lexeme, "printn") == 0) token->code = codePRINTN;
        break;
    }
  }

  /*REMOVED (10-18-02)
  // Mark type as identifier //
  if (token->code == codeIdentifier)
  {
    token->type = tIdentifier;
  }*/

  return token;
}

//// mGetNumber
Token *mGetNumber ()
{
  // Variables //
  Token *token;
  int nPos;
  int nMantissaLen;

  token = tok_new();
  nPos = 0;
  nMantissaLen = -1; // Assume the token is an integer

  token->lexeme[nPos] = ib_get ();
  ib_seek ();

  while (k_inSet (ib_get(), LEXER_NUMBER_ELEMENTS))
  {
    if (nMantissaLen > -1) // if currently in mantissa, increase mantissa length
    {
      nMantissaLen++;
    }
    if (ib_get() == '.')
    {
      nMantissaLen++;
    }
    token->lexeme[++nPos] = ib_get();
    ib_seek ();
  }

  // Fill any empty mantissa places with zeros //
  if (nMantissaLen > -1)
  {
    while (nMantissaLen < GLOBAL_MANTISSA_SIZE)
    {
      nMantissaLen++;
      token->lexeme[++nPos] = '0';
    }
  }

  // End lexeme string //
  token->lexeme[++nPos] = 0;

  if (nMantissaLen > -1) // rational number
  {
    token->code = codeRatLiteral;
    tok_set (token, k_new_f(conv_sf(token->lexeme)));
  }
  else // integer number
  {
    token->code = codeIntLiteral;
    tok_set (token, k_new_n(conv_si(token->lexeme)));
  }

  return token;
}

//// mGetString
Token *mGetString ()
{
  // Variables //
  Token *token;
  char c;
  int pos;

  token = tok_new();
  pos = 0;

  // Token is a string literal //
  token->code = codeStrLiteral;

  // Opening quote //
  token->lexeme[pos++] = '\"';

  // Get the string //
  ib_seek (); // skip opening quote
  c = ib_get ();
  while (c != INPUTBUFFER_EOF)
  {
    if (c == '\"') // current char is quote
    {
      ib_seek ();
      c = ib_get ();
      if (c != '\"')
      {
        break; // not a double quote, so end the string
      }
    }
    else if (c == 0) // end of line reached
    {
      c = ' '; // replace end-line with blank
    }

    // Append current char to string //
    token->lexeme[pos++] = c;

    // Get next char //
    ib_seek ();
    c = ib_get ();
  }

  // BUG?:
  // Should possibly be removed, since the first char after a string may be eof
  if (c == INPUTBUFFER_EOF)
  {
    err (true, EF_UNEXPECTED_EOF, NULL);
    erre (EFF_IGNORE);
  }

  token->lexeme[pos++] = '\"'; // closing quote
  token->lexeme[pos] = '\0';

  // After this, token->lexeme will contain the string with quotes and //
  // token->value will contain the string minus the quotes             //
  tok_set (token, k_sub (token->lexeme, 1, strlen (token->lexeme) - 1));

  return token;
}

//// mGetEOL
Token *mGetEOL ()
{
  Token *token = tok_new();

  token->code = codeEOL;
  token->lexeme[0] = 0;

  ib_seek ();

  return token;
}

// mGetSpecial
Token *mGetSpecial ()
{
  // Variables //
  Token *token;
  char c;
  int pos;

  token = tok_new();
  pos = 0;

  c = ib_get ();
  token->lexeme[pos++] = c;

  switch (c) {
    case '+': token->code = codePLUS;         ib_seek (); break;
    case '-': token->code = codeMINUS;        ib_seek (); break;
    case '*': token->code = codeTIMES;        ib_seek (); break;
    case '/': token->code = codeFORWARDSLASH; ib_seek (); break;
    case '\\': token->code = codeBACKSLASH;   ib_seek (); break;
    case '=': token->code = codeEQUALS;       ib_seek (); break;
    case '(': token->code = codeLEFTPAREN;    ib_seek (); break;
    case ')': token->code = codeRIGHTPAREN;   ib_seek (); break;
    case ',': token->code = codeCOMMA;        ib_seek (); break;
    case '$': token->code = codeDOLLAR;       ib_seek (); break;
    case '.': token->code = codeDOT;          ib_seek (); break;
    case '[': token->code = codeLEFTBRACK;    ib_seek (); break;
    case ']': token->code = codeRIGHTBRACK;   ib_seek (); break;
    case '\"': token->code = codeQUOTEMARK;   ib_seek (); break;
    case ';': token->code = codeSEMICOLON;    ib_seek (); break;
    case ':': token->code = codeCOLON;        ib_seek (); break;
    case '&': token->code = codeANDSIGN;      ib_seek (); break;
    case '#': token->code = codePOUNDSIGN;    ib_seek (); break;
    case '<': // < or <= or <> or <-
              ib_seek ();
              c = ib_get ();
              if (c == '=') // <=
              {
                token->lexeme[pos++] = '=';
                token->code = codeLTEQUAL;
                ib_seek ();
              }
              else if (c == '>') // <>
              {
                token->lexeme[pos++] = '>';
                token->code = codeNOTEQUAL;
                ib_seek ();
              }
              else if (c == '-') // <-
              {
                token->lexeme[pos++] = '-';
                token->code = codeARROW;
                ib_seek ();
              }
              else // <
              {
                token->code = codeLT;
              }
              break;
    case '>': // > or >=
              ib_seek ();
              c = ib_get ();
              if (c == '=') // >=
              {
                token->lexeme[pos++] = '=';
                token->code = codeGTEQUAL;
                ib_seek ();
              }
              else // >
              {
                token->code = codeGT;
              }
              break;
    default: // error
             token->code = codeNull;
             ib_seek ();
             token->lexeme[pos] = '\0'; // must be here so that lexeme can be reported
             err (true, EF_UNDEFINED_TOKEN, token->lexeme);
             erre (EFF_IGNORE);
             break;
  }
  token->lexeme[pos] = '\0';

  return token;
}

//// mGetEOF
Token *mGetEOF ()
{
  Token *token = tok_new();

  token->code = codeEOF;
  token->lexeme[0] = 0;

  return token;
}

//// mGetError
Token *mGetError ()
{
  Token *token = tok_new();

  token->code = codeError;
  token->lexeme[0] = ib_get ();
  token->lexeme[1] = '\0';

  ib_seek ();
  err (true, EF_UNDEFINED_TOKEN, token->lexeme);
  erre (EFF_IGNORE);

  return token;
}

#endif
